/***
   This file is part of fly-ocr.

   Copyright (C) 2016-2020 Sandro Mani <manisandro@gmail.com>
   Copyright (C) 2021 LLC "RusBITech-Astra" (https://astralinux.ru)

    Based on gImageReader source code by Sandro Mani under the GNU
    General Public License (GPL) version 3 or later

   Foobar is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   Foobar is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with fly-ocr. If not, see <https://www.gnu.org/licenses/>.
***/

#include "displayertoolselect.h"
#include "displayer.h"
#include "filedialogs.h"
#include "mainwindow.h"
#include "recognizer.h"
#include "utils.h"
#include <cmath>
#include <QHBoxLayout>
#include <QLabel>
#include <QMenu>
#include <QMouseEvent>
#include <QStyle>
#include <fly-ocr/ocrengine.h>

DisplayerToolSelect::DisplayerToolSelect(Displayer* displayer, QObject* parent)
    : DisplayerTool(displayer, parent)
{
    displayer->setCursor(Qt::CrossCursor);
    updateRecognitionModeLabel();
}

DisplayerToolSelect::~DisplayerToolSelect()
{
    clearSelections();
}

void DisplayerToolSelect::mousePressEvent(QMouseEvent* event)
{
    if (event->button() == Qt::LeftButton &&  m_curSel == nullptr) {
        if ((event->modifiers() & Qt::ControlModifier) == 0) {
            clearSelections();
        }
        m_curSel = new NumberedDisplayerSelection(this, 1 + m_selections.size(), m_displayer->mapToSceneClamped(event->pos()));
        m_curSel->setZValue(1 + m_selections.size());
        m_displayer->scene()->addItem(m_curSel);
        event->accept();
    }
}

void DisplayerToolSelect::mouseMoveEvent(QMouseEvent* event)
{
    if (m_curSel) {
        QPointF p = m_displayer->mapToSceneClamped(event->pos());
        m_curSel->setPoint(p);
        m_displayer->ensureVisible(QRectF(p, p));
        event->accept();
    }
}

void DisplayerToolSelect::mouseReleaseEvent(QMouseEvent* event)
{
    if (m_curSel) {
        if (m_curSel->rect().width() < 5.0 || m_curSel->rect().height() < 5.0) {
            delete m_curSel;
        } else {
            m_selections.append(m_curSel);
            updateRecognitionModeLabel();
        }
        m_curSel = nullptr;
        event->accept();
    }
}

void DisplayerToolSelect::resolutionChanged(double factor)
{
    for (NumberedDisplayerSelection* sel : m_selections) {
        sel->scale(factor);
    }
}

void DisplayerToolSelect::rotationChanged(double delta)
{
    QTransform t;
    t.rotate(delta);
    for (NumberedDisplayerSelection* sel : m_selections) {
        sel->rotate(t);
    }
}

QList<QImage> DisplayerToolSelect::getOCRAreas()
{
    QList<QImage> images;
    if (m_selections.empty()) {
        images.append(m_displayer->getImage(m_displayer->getSceneBoundingRect()));
    } else {
        for (const NumberedDisplayerSelection* sel : m_selections) {
            images.append(m_displayer->getImage(sel->rect()));
        }
    }
    return images;
}

void DisplayerToolSelect::clearSelections()
{
    qDeleteAll(m_selections);
    m_selections.clear();
    updateRecognitionModeLabel();
}

void DisplayerToolSelect::removeSelection(int num)
{
    delete m_selections[num - 1];
    m_selections.removeAt(num - 1);
    for (int i = 0, n = m_selections.size(); i < n; ++i) {
        m_selections[i]->setNumber(1 + i);
        m_selections[i]->setZValue(1 + i);
    }
}

void DisplayerToolSelect::reorderSelection(int oldNum, int newNum)
{
    NumberedDisplayerSelection* sel = m_selections[oldNum - 1];
    m_selections.removeAt(oldNum - 1);
    m_selections.insert(newNum - 1, sel);
    for (int i = 0, n = m_selections.size(); i < n; ++i) {
        m_selections[i]->setNumber(1 + i);
        m_selections[i]->setZValue(1 + i);
    }
}

void DisplayerToolSelect::saveSelection(NumberedDisplayerSelection* selection)
{
    QImage img = m_displayer->getImage(selection->rect());
    QString filename = FileDialogs::saveDialog(tr("Save Selection Image"), tr("selection.png"), "outputdir", QString("%1 (*.png)").arg(tr("PNG Images")), true);
    if (!filename.isEmpty()) {
        img.save(filename);
    }
}

void DisplayerToolSelect::updateRecognitionModeLabel()
{
    MAIN->recognizer()->setRecognizeMode(m_selections.isEmpty() ? tr("Recognize all") : tr("Recognize selection"));
}

void DisplayerToolSelect::autodetectLayout(bool noDeskew)
{
    clearSelections();

    double avgDeskew = 0.0;
    QList<QRectF> rects;
    QImage img = m_displayer->getImage(m_displayer->getSceneBoundingRect());

    OcrEngine* engine = OcrEngine::create(this, QStringList(), true);
    if (!engine) {
        qWarning("DisplayerToolSelect: unable to create OCR engine");
        return;
    }

    // Perform layout analysis
    Utils::busyTask([engine, &avgDeskew, &rects, &img] {
        rects = engine->analyseLayout(img, &avgDeskew);
        return true;
    }, tr("Performing layout analysis"));

    // If a somewhat large deskew angle is detected, automatically rotate image and redetect layout,
    // unless we already attempted to rotate (to prevent endless loops)
    if (std::abs(avgDeskew) > 0.1 && !noDeskew) {
        double newangle = m_displayer->currentAngle() - avgDeskew;
        m_displayer->setup(nullptr, nullptr, &newangle);
        autodetectLayout(true);
    } else {
        // Merge overlapping rectangles
        for (int i = rects.size(); i-- > 1;) {
            for (int j = i; j-- > 0;) {
                if (rects[j].intersects(rects[i])) {
                    rects[j] = rects[j].united(rects[i]);
                    rects.removeAt(i);
                    break;
                }
            }
        }
        for (int i = 0, n = rects.size(); i < n; ++i) {
            m_selections.append(new NumberedDisplayerSelection(this, 1 + i, rects[i].topLeft()));
            m_selections.back()->setPoint(rects[i].bottomRight());
            m_displayer->scene()->addItem(m_selections.back());
        }
        updateRecognitionModeLabel();
    }

    delete engine;
}

///////////////////////////////////////////////////////////////////////////////

void NumberedDisplayerSelection::contextMenuEvent(QGraphicsSceneContextMenuEvent* event)
{
    QMenu menu;

    QWidget* orderWidget = new QWidget(&menu);
    QHBoxLayout* layout = new QHBoxLayout(orderWidget);

    QLabel* orderIcon = new QLabel(&menu);
    int iconSize = orderIcon->style()->pixelMetric(QStyle::PM_SmallIconSize);
    orderIcon->setPixmap(QIcon::fromTheme("object-order-front").pixmap(iconSize, iconSize));
    layout->addWidget(orderIcon);
    layout->setContentsMargins(4, 0, 4, 0);

    QLabel* orderLabel = new QLabel(tr("Order:"), &menu);
    orderLabel->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Fixed);
    layout->addWidget(orderLabel);

    QSpinBox* orderSpin = new QSpinBox();
    orderSpin->setRange(1, static_cast<DisplayerToolSelect*>(m_tool)->m_selections.size());
    orderSpin->setValue(m_number);
    orderSpin->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Fixed);
    connect(orderSpin, qOverload<int>(&QSpinBox::valueChanged), this, &NumberedDisplayerSelection::reorderSelection);
    layout->addWidget(orderSpin);

    QWidgetAction* spinAction = new QWidgetAction(&menu);
    spinAction->setDefaultWidget(orderWidget);

    QAction* deleteAction = new QAction(QIcon::fromTheme("edit-delete"), tr("Delete"), &menu);
    QAction* ocrAction = new QAction(QIcon::fromTheme("insert-text"), tr("Recognize"), &menu);
    QAction* ocrClipboardAction = new QAction(QIcon::fromTheme("edit-copy"), tr("Recognize to clipboard"), &menu);
    QAction* saveAction = new QAction(QIcon::fromTheme("document-save-as"), tr("Save as image"), &menu);
    menu.addActions(QList<QAction*>() << spinAction << deleteAction << ocrAction << ocrClipboardAction << saveAction);
    QAction* selected = menu.exec(event->screenPos());
    if (selected == deleteAction) {
        static_cast<DisplayerToolSelect*>(m_tool)->removeSelection(m_number);
    } else if (selected == ocrAction) {
        MAIN->recognizer()->recognizeImage(m_tool->displayer()->getImage(rect()), Recognizer::OutputDestination::Buffer);
    } else if (selected == ocrClipboardAction) {
        MAIN->recognizer()->recognizeImage(m_tool->displayer()->getImage(rect()), Recognizer::OutputDestination::Clipboard);
    } else if (selected == saveAction) {
        static_cast<DisplayerToolSelect*>(m_tool)->saveSelection(this);
    }
}

void NumberedDisplayerSelection::reorderSelection(int newNumber)
{
    static_cast<DisplayerToolSelect*>(m_tool)->reorderSelection(m_number, newNumber);
}

void NumberedDisplayerSelection::paint(QPainter* painter, const QStyleOptionGraphicsItem* option, QWidget* widget)
{
    DisplayerSelection::paint(painter, option, widget);

    painter->setRenderHint(QPainter::Antialiasing, false);
    QRectF r = rect();
    qreal w = 20.0 / m_tool->displayer()->currentScale();
    w = std::min(w, std::min(r.width(), r.height()));
    QRectF box(r.x(), r.y(), w, w);
    painter->setBrush(QPalette().highlight());
    painter->drawRect(box);
    painter->setRenderHint(QPainter::Antialiasing, true);

    if (w > 1.25) {
        QFont font;
        font.setPixelSize(0.8 * w);
        font.setBold(true);
        painter->setFont(font);
        painter->setPen(QPalette().highlightedText().color());
        painter->drawText(box, Qt::AlignCenter, QString::number(m_number));
    }
}
